static char *CopyrightIdentifier(void) { return "@(#)pgmcp.cc 0.12 95/09/18 Copyright (C) 1993,1994,1995. David A. Clunie. All rights reserved."; }

#include <iostream.h>
// #include <iomanip.h>
#include <strstream.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#include "GetOpt.h"
#include "MainCmd.h"

static
char
pnm_getc(istream &istr)		// 1st non-comment, non-whitespace, 0 on failure
{
	char c;
	while (istr.get(c) && (c == ' ' || c == '\t' || c == '\n'));
	if (!istr) return 0;
	if (c == '#') {		// skip comments
		while (istr.get(c) && c != '\n');
		if (!istr) return 0;
		while (istr.get(c) && (c == ' ' || c == '\t' || c == '\n'));
		if (!istr) return 0;
	}
	return c;
}

static
int
pnm_getu(istream &istr)		// get unsigned value, -1 on failure
{
	char c=pnm_getc(istr);	// skips comments and whitespace
	if(!c || !isdigit(c)) return -1;
	unsigned value=0;
	while (c && isdigit(c)) {
		value*=10;
		value+=(c-'0');
		istr.get(c);
	}
	// discards single trailing white space
	return value;
}

static
unsigned short
pnm_geti(istream &istr,unsigned short wordlength)	// get image data value
{
	unsigned char c[2];
	istr.read(c,wordlength);
	unsigned short value;
	// PNM format is little-endian ...
	if (wordlength == 2)
		value=(c[1]<<8)+c[0];
	else
		value=c[0];
	return value;
}

static
void
pnm_puti(ostream &ostr,unsigned short value,unsigned short wordlength)
							// put image data value
{
	unsigned char c[2];
	// PNM format is little-endian ...
	if (wordlength == 2) {
		c[0]=value&0xff;
		c[1]=value>>8;
	}
	else
		c[0]=value;
	ostr.write(c,wordlength);
}

int
main(int argc, char *argv[])
{
	CCOMMAND(argc,argv);

	cerr.setf(ios::showbase,ios::showbase);

	int		verbose = 0;
	char		filter = 0;
	char		unfilter = 0;
	unsigned databits	= 0;

	GetOpt getopt(argc,argv,"d:f:u:v");
	int c;
	int err=0;
	while ((c=getopt()) != -1 && !err) {
		switch (c) {
			case 'v':
				verbose=1;
				break;
			case 'd':
				databits=atoi(getopt.optarg);
				break;
			case 'f':	// Filter method
				switch (getopt.optarg[0]) {
					case 's':	// Subtract
					case 'S':
						filter='s';
						break;
					case 'c':	// Cross
					case 'C':
						filter='c';
						break;
					case 'n':	// None
					case 'N':
						filter=0;
						break;
					default:
						cerr << "Bad filter type\n"
						     << flush;
						err=1;
						break;
				}
				break;
			case 'u':	// Unfilter method
				switch (getopt.optarg[0]) {
					case 's':	// Subtract
					case 'S':
						unfilter='s';
						break;
					case 'c':	// Cross
					case 'C':
						unfilter='c';
						break;
					case 'n':	// None
					case 'N':
						unfilter=0;
						break;
					default:
						cerr << "Bad unfilter type\n"
						     << flush;
						err=1;
						break;
				}
				break;
			case '?':
			default:
				err=1;
				break;
		}
	}

	if (getopt.optind != argc) {
		cerr << "Trailing garbage after options\n" << flush;
		err=1;
	}

	if (databits && databits<8) databits*=8;

	if (databits && databits%8) {
		cerr << "Error - data length must be 1,2 or multiple of 8\n"
		     << flush;
		err=1;
	}

	if (err) {
		cerr << "Usage: " << argv[0]
		     << " [-d databits]"
		     << " [-f subtract|cross|none]"
		     << " [-u subtract|cross|none] [-v]\n"
		     << flush;
		return 1;
	}

	// Read PNM header ...

	char magic=pnm_getc(cin);
	char format;
	cin.get(format);

	if (!cin || magic != 'P') {
		cerr << "Not pnm file\n" << flush;
		return 1;
	}

	if (format != '5') {
		cerr << "Unsupported pnm format P" << format << "\n" << flush;
		return 1;
	}

	int width = pnm_getu(cin);
	int height = pnm_getu(cin);
	int maxval = pnm_getu(cin);

	if (verbose) {
		cerr << "width " << width << "\n" << flush;
		cerr << "height " << height << "\n" << flush;
		cerr << "maxval " << maxval << "\n" << flush;
		if (filter) cerr << "filter " << filter 
				   << "\n" << flush;
		if (unfilter) cerr << "unfilter " << unfilter 
				     << "\n" << flush;
	}

	if (!cin || width == -1 || height == -1 || maxval == -1) {
		cerr << "Corrupt pnm header\n" << flush;
		return 1;
	}

	// Last whitespace discarded ... at start of data

	// If -d is specified, apply the filters as if the word length
	// was that value, rather than what it actually is based on
	// maxval.

	unsigned wordlength;
	unsigned maxvallength=(maxval <= 255) ? 1 : 2;
	unsigned widthfactor;
	if (databits) {
	 	wordlength=(databits-1)/8u+1;
		if ((widthfactor=maxvallength/wordlength) == 0) {
			cerr << "Error - pbm word length < -d length\n"<< flush;
			return 1;
		}
	}
	else {
	 	wordlength=maxvallength;
		widthfactor=1;
	}

	unsigned correctedwidth=width*widthfactor;

	// Write new PNM header ...

	{
		ostrstream ost;
		ost << "P5\n" << width << " " << height << "\n"
		    << maxval << "\n" << '\0';
		char *str=ost.str();
		cout.write(str,strlen(str));
		delete str;
	}

	// Copy image data ...

	unsigned short *clastrowm1;
	unsigned short *clastrow;
	unsigned short *dlastrowm1;
	unsigned short *dlastrow;
	unsigned short *clastpixel;
	unsigned short *dlastpixel;

	int x;
	unsigned short *ptr;
	if (filter == 'c') {
		clastrowm1 = new unsigned short[correctedwidth+widthfactor];
		clastrow   = clastrowm1+widthfactor;
		for (x=-widthfactor,ptr=clastrowm1; x<correctedwidth; ++x) *ptr++=0;
		clastpixel = new unsigned short[widthfactor];
	}
	if (unfilter == 'c') {
		dlastrowm1 = new unsigned short[correctedwidth+1];
		dlastrow   = dlastrowm1+widthfactor;
		for (x=widthfactor,ptr=dlastrowm1; x<correctedwidth; ++x) *ptr++=0;
		dlastpixel = new unsigned short[widthfactor];
	}

	unsigned y;
	unsigned short mask=(wordlength==2)?0xffff:0xff;
	for (y=0; y<height; ++y) {
		if (filter == 'c') {
			unsigned p;
			for (p=0; p<widthfactor; ++p) clastpixel[p]=0;
		}
		if (unfilter == 'c') {
			unsigned p;
			for (p=0; p<widthfactor; ++p) dlastpixel[p]=0;
		}
		for (x=0; x<correctedwidth; x+=widthfactor) {
			unsigned p;
			for (p=0; p<widthfactor; ++p) {
				unsigned short value=pnm_geti(cin,wordlength);

				if (unfilter == 's') {
					value=(value+dlastpixel[p])&mask;
					dlastpixel[p]=value;
				}
				else if (unfilter == 'c') {
					value=(value
					      +dlastpixel[p]
					      +dlastrow[x+p]
					      -dlastrow[x-widthfactor+p]
					      )&mask;
					dlastrow[x-widthfactor+p]=dlastpixel[p];
					dlastpixel[p]=value;
				}
				if (filter == 's') {
					unsigned short temp;
					temp=(value-clastpixel[p])&mask;
					clastpixel[p]=value;
					value=temp;
				}

				else if (filter == 'c') {
					unsigned short temp;
					temp=(value
						 -clastpixel[p]
						 -clastrow[x+p]
						 +clastrow[x-widthfactor+p]
						 )&mask;
					clastrow[x-widthfactor+p]=clastpixel[p];
					clastpixel[p]=value;
					value=temp;
				}

				pnm_puti(cout,value,wordlength);
			}
		}
		unsigned p;
		for (p=0; p<widthfactor; ++p) {
			if (filter == 'c') 
				clastrow[correctedwidth-widthfactor+p]=clastpixel[p];
			if (unfilter == 'c') 
				dlastrow[correctedwidth-widthfactor+p]=dlastpixel[p];
		}
	}

	if (!cin.good()) {
		cerr << "Read error\n" << flush;
		return 1;
	}
	if (!cout.good()) {
		cerr << "Write error\n" << flush;
		return 1;
	}

	return 0;
}

